from bs4 import element
from matplotlib.cm import ScalarMappable, RdYlGn

from selectors import *
from actors import *
from evaluators import *

from numpy import log


def score_to_rgb(score, sm, min_score, max_score):
    """Only used for debugging"""

    if score > 0:
        score = log(score)
        val = 0.5 + ((score/float(max_score))/2.5)
    elif score < 0:
        score = -log(-score)
        val = 0.5 - ((score/float(min_score))/2.5)
    else:
        val = 0.5
    r, g, b, a = sm.to_rgba(val)
    return (int(r*255), int(g*255), int(b*255))


def style_all_tags(soup, container, background=True, margin=None, border=False, css=True, images=True, only_container=False):
    """Only used for debugging"""

    #if only_container:
        #soup = container

    (min_score, max_score) = get_min_and_max_scores(soup)
    min_score = -log(-min_score)
    max_score = log(max_score)

    sm = ScalarMappable(cmap=RdYlGn)
    sm.set_clim(0,1)

    if not css:
        for link in soup.find_all("link"):
            link.decompose()

    if not images:
        for img in soup.find_all("img"):
            img.decompose()

    for tag in soup.find_all():

        if not isinstance(tag, element.Tag):
            continue

        if css:
            try:
                style = tag["style"].split(";")
            except:
                style = []
        else:
            style = []

        if margin is not None:
            if margin > 0:
                style.append("margin: %dpx" % margin)

        try:
            score = sum(tag.scores.values())
            rgb = score_to_rgb(score, sm, min_score, max_score)
            tag.scores['total'] = score
        except:
            score = -1
            rgb = (230, 230, 230)

        if background:
            style.append("background-color: rgb(%d,%d,%d)" % rgb)

        try:
            if tag in container.descendants and score > -30:
                style.append("color: #000000")
            else:
                style.append("color: #666666")
        except:
            pass

        if tag == container:
            style.append("border: 3px dashed #0000CC")
            style.append("color: #000000")
        else:
            if border:
                style.append("border: 1px solid #333333")

        try:
            tag['style'] = "; ".join(style)
            tag['scores'] = repr(tag.scores)
            del tag.scores['total']
        except:
            pass


def get_min_and_max_scores(tag):

    min_score = -10
    max_score = 10

    for child in tag.find_all():
        score = get_total_score(child)
        min_score = min(min_score, score)
        max_score = max(max_score, score)

    return (min_score, max_score)


def get_tag_with_max_score(tag):

    max_child = None
    max_score = -10000

    for child in tag.find_all():
        if child is not None:
            score = get_total_score(child)
            if score >= max_score:
                max_score = score
                max_child = child

    return max_child


def get_total_score(tag):
    try:
        return sum(tag.scores.values())
    except:
        return 0



from bs4 import element

from core import *
from selectors import *
from evaluators import *

class Actor(object):
    """Acts on one tag"""

    def __init__(self, evaluator=None):
        self.evaluator = evaluator

    def act(self, tag):
        pass


class Scorer(Actor):

    def __init__(self, evaluator, reset_children=False):
        Actor.__init__(self, evaluator)
        self.reset_children = reset_children

    def act(self, tag):

        if tag.scores is None:
            tag.scores = dict()
        tag.scores[self.evaluator.name] = self.evaluator.evaluate(tag)

        if self.reset_children:
            for child in tag.descendants:
                try:
                    del child.scores[self.evaluator.name]
                except:
                    pass


class Pruner(Actor):

    def __init__(self, evaluator=None):
        Actor.__init__(self, evaluator)

    def act(self, tag):
        if self.evaluator is not None:
            if self.evaluator.evaluate(tag) >= 1:
                try:
                    tag.decompose()
                except:
                    tag.replace_with("")

        else:
            try:
                tag.decompose()
            except:
                tag.replace_with("")


class RegExReplacer(Actor):

    def __init__(self, pattern="", repl=""):
        super(RegExReplacer, self).__init__(self)

        if isinstance(pattern, str):
            self.pattern = re.compile(pattern, flags=re.UNICODE)
        else:
            self.pattern = pattern

        self.repl = repl

    def act(self, tag):

        if isinstance(tag, element.Tag):
            tag.string = self.replace(tag.get_text())
        else:
            text = tag
            return self.replace(text)

    def replace(self, text):
        return self.pattern.sub(self.repl, text)



#!/usr/bin/python2.7

import sys
import re
import argparse
import fileinput
import requests
import codecs
from os import path, listdir, makedirs
from copy import deepcopy

from bs4 import BeautifulSoup, Comment, NavigableString
import urlparse

from collections import Counter

from autocues import *
from core import get_tag_with_max_score, get_total_score, style_all_tags


class Reporter(object):

    news_container = None

    def read(self, url=None, html=None, soup=None, autocue=default_autocue):

        self.autocue = autocue

        if soup is not None:
            self.soup = soup
        else:
            if html is not None:
                self.soup = self._get_soup(html)
            else:
                if url is not None:
                    html = requests.get(url).content
                    self.soup = self._get_soup(html)
                else:
                    self.soup = None

        self.make_urls_absolute(self.soup, url)

        self.autocue.execute(self.soup, PRE_TRAVERSAL)

        # We work our way up the DOM
        for tag in reversed(self.soup.find_all()):

            if tag.name == 'p':
                evaluate_as = EVAL_PARAGRAPH
            else:
                # If tag contains text of its own, evaluate it as a paragraph
                evaluate_as = EVAL_CONTAINER
                for child in tag.children:
                    if isinstance(child, NavigableString):
                        text = unicode(child).strip()
                        if len(text) > 10:
                            evaluate_as = EVAL_PARAGRAPH
                            continue

            self.autocue.execute(tag, evaluate_as)

        self.autocue.execute(self.soup, POST_TRAVERSAL)
        self.news_container = get_tag_with_max_score(self.soup)


    def report_styled_soup(self, background=True, margin=None, border=False, css=True, images=True, only_container=False):


        if only_container:
            self.soup = self.news_container

        style_all_tags(self.soup, self.news_container, background, margin, border, css, images, only_container)

        return self.soup


    def report_news(self):

        self.autocue.execute(self.news_container, NEWS_CONTAINER)
        text = self.news_container.get_text()
        text = self.autocue.execute(text, NEWS_TEXT)
        return text


    def _get_soup(self, html):

        html = self.autocue.execute(html, HTML)

        # Create a soup using the lxml parser
        soup = BeautifulSoup(html, "lxml")


        return soup

    def make_urls_absolute(self, soup, url):

        if url is None:
            url = self.guess_base_url(soup)


        for tag in soup.findAll('a', href=True):
            tag['href'] = urlparse.urljoin(url, tag['href'])

        for tag in soup.findAll('link', href=True):
            tag['href'] = urlparse.urljoin(url, tag['href'])

        for tag in soup.findAll('img', href=True):
            tag['src'] = urlparse.urljoin(url, tag['src'])


    def guess_base_url(self,soup):

        urls = []
        base_urls = []
        urls.extend([tag['href'] for tag in soup.findAll('a', href=True)])
        urls.extend([tag['href'] for tag in soup.findAll('link', href=True)])
        urls.extend([tag['src'] for tag in soup.findAll('img', href=True)])

        for url in urls:
            res = urlparse.urlparse(url)
            base_urls.append(res.scheme + "://" + res.netloc + "/")

        try:
            cnt = Counter(base_urls)
            url = cnt.most_common(1)[0][0]
        except IndexError:
            url = ""

        return url

def main():

    parser = argparse.ArgumentParser(description="Extract news from an HTML file", epilog="Specify either a url (--url) or a list of input files (--input).")
    parser.add_argument('--url', type=str)
    parser.add_argument('--input', metavar='FILE', nargs='+', type=str)
    parser.add_argument('--debug', action='store_true', help="Style the HTML based on scores and save to outputXX.html")
    parser.add_argument('--test', action='store_true', help="process every HTML file in test/input and store the output in test/ouput/")
    args = parser.parse_args()

    if (not args.test) and args.url is None and args.input is None:
        parser.error("Specify either --test, --url or --input.")

    my_reporter = Reporter()


    if args.test:
        test_path = 'test'

        for dirname in ['html_01','html_02','html_03','html_04','html_05','text']:
            try:
                makedirs(path.join(test_path, 'output', dirname))
            except:
                pass

        for name in sorted(listdir(path.join(test_path, 'input'))):

            print "Processing %s ..." % name

            if path.isfile(path.join(test_path, 'output', 'html_01', name)):
                continue


            #with codecs.open(path.join(test_path, 'input', name), 'r', encoding='utf-8') as f:


            with open(path.join(test_path, 'input', name), 'rb') as f:
                my_reporter.read(html=f.read())


            with codecs.open(path.join(test_path, 'output', 'html_01', name), 'w+', encoding='utf-8') as f:
                f.write(my_reporter.report_styled_soup(background=False).prettify(formatter="html"))

            with codecs.open(path.join(test_path, 'output', 'html_02', name), 'w+', encoding='utf-8') as f:
                f.write(my_reporter.report_styled_soup(background=True).prettify(formatter="html"))

            with codecs.open(path.join(test_path, 'output', 'html_03', name), 'w+', encoding='utf-8') as f:
                f.write(my_reporter.report_styled_soup(background=True, margin=2).prettify(formatter="html"))

            with codecs.open(path.join(test_path, 'output', 'html_04', name), 'w+', encoding='utf-8') as f:
                f.write(my_reporter.report_styled_soup(background=True, margin=5, border=True, css=False, images=False).prettify(formatter="html"))

            with codecs.open(path.join(test_path, 'output', 'text', path.splitext(name)[0]+'.txt'), 'w+', encoding='utf-8') as f:
                f.write(my_reporter.report_news())

            with codecs.open(path.join(test_path, 'output', 'html_05', name), 'w+', encoding='utf-8') as f:
                f.write(my_reporter.report_styled_soup(background=True, margin=5, border=True, css=False, images=False, only_container=True).prettify(formatter="html"))



    else:
        debug_counter = 0
        if args.url is not None:
            my_reporter.read(url=args.url)

            if args.debug:
                debug_counter += 1
                write_debug(my_reporter, debug_counter)
            else:
                print my_reporter.report_news()

        else:
            for filename in args.input:
                with codecs.open(filename, 'r', encoding='utf-8') as f:
                    my_reporter.read(html=f.read())

                    if args.debug:
                        debug_counter += 1
                        write_debug(my_reporter, debug_counter)
                    else:
                        print my_reporter.report_news()


def write_debug(reporter, debug_counter):

    with codecs.open('debug/output%02d.html' % debug_counter, 'w+', encoding='utf-8') as f:
        f.write(reporter.report_styled_soup(background=True).prettify(formatter="html"))

    with codecs.open('debug/output%02d.txt' % debug_counter, 'w+', encoding='utf-8') as f:
        f.write(reporter.report_news())


if __name__ == '__main__':
    sys.exit(main())



from bs4 import element, Comment

from core import *
from actors import *
from evaluators import *
import core

class Selector(object):
    """Selects one or more tags and then calls the Actor"""

    def __init__(self, actor=None):
        self.actor = actor

    def select(self, tag):
        pass


class SingleTagSelector(Selector):

    def __init__(self, actor=None):
        Selector.__init__(self, actor)

    def select(self, tag):
        if isinstance(tag, element.Tag):
            self.actor.act(tag)


class CommentSelector(Selector):

    def __init__(self, actor=None):
        Selector.__init__(self, actor)

    def select(self, tag):
        for child in reversed(tag.find_all(text=lambda text:isinstance(text, Comment))):
            self.actor.act(child)


class ScoreSelector(Selector):

    def __init__(self, threshold=0, mode="upper", actor=None):
        Selector.__init__(self, actor)

        self.threshold = threshold
        self.mode = mode

    def select(self, tag):
        if isinstance(tag, element.Tag):
            for child in reversed(tag.find_all()):
                if not isinstance(tag, element.Tag):
                    continue
                score = core.get_total_score(child)

                if self.mode == "upper":
                    if score < self.threshold:
                        self.actor.act(child)
                else:
                    if score > self.threshold:
                        self.actor.act(child)


class CSSSelector(Selector):

    def __init__(self, select_string, actor=None):

        Selector.__init__(self, actor)
        self.select_string = select_string

    def select(self, tag):

        for child in reversed(tag.select(self.select_string)):
            if isinstance(tag, element.Tag):
                self.actor.act(child)


class BeautifulSelector(Selector):

    def __init__(self, actor=None, **kwargs):

        Selector.__init__(self, actor)
        self.args = kwargs

    def select(self, tag):

        for child in reversed(tag.find_all(**self.args)):
            if isinstance(tag, element.Tag):
                self.actor.act(child)



from bs4 import BeautifulSoup, element
import re
from matplotlib.cm import ScalarMappable, RdYlGn

from core import *
from selectors import *
from actors import *
from evaluators import *
from autocues import *

# Triggers:








policy = ScoringPolicy()

policy.append((ScoreAggregator(), EVAL_CONTAINER))
policy.append((RegExCounter(",", 1), EVAL_PARAGRAPH))
policy.append((Limiter(RegExCounter("(\w)+(['`]\w)?", 1), 20), EVAL_PARAGRAPH))
policy.append((CSSSelector("a", Scorer(FixedValue(-50))), PRE_TRAVERSAL))
policy.append((ScoreSelector(Pruner()), POST_TRAVERSAL))

#policy['parent'] = (ParentScorer(), EVAL_CONTAINER) # aggregates scores from children
#policy['commas'] = (RegExCountScorer(",", 1), EVAL_PARAGRAPH)
#policy['words'] = (RegExCountScorer("(\w)+(['`]\w)?", 1), EVAL_PARAGRAPH)
#policy['children'] = (CountChildrenScorer("p", 10), EVAL_CONTAINER)
#policy['paragraph'] = (FixedValueScorer(1), EVAL_PARAGRAPH)
#policy['container'] = (FixedValueScorer(1), EVAL_CONTAINER)

if __name__ == '__main__':

    html_doc = """ <html><head><title>The Dormouse's story</title></head><body><div><p class="title"><b>The Dormouse's story</b></p> <p class="story">Once upon a time there were three little sisters; and their names were <a href="http://example.com/elsie" class="sister" id="link1">Elsie</a>, <a href="http://example.com/lacie" class="sister" id="link2">Lacie</a> and <a href="http://example.com/tillie" class="sister" id="link3">Tillie</a>; and they lived at the bottom of a well.</p> <p class="story">...</p></div></body></html>"""

    soup = BeautifulSoup(html_doc)

    for tag in reversed(soup.find_all(["p","a","b"])):
        policy.execute(tag, EVAL_PARAGRAPH)

    policy.execute(soup, PRE_TRAVERSAL)

    policy.execute(soup.find("body"), EVAL_CONTAINER)

    for tag in soup.find_all():
        print tag.name, "=", tag.scores

    container = get_tag_with_max_score(soup)

    style_all_tags(soup, container)

    html_styled = soup.find("body").prettify(formatter="html")

    policy.execute(container, POST_TRAVERSAL)
    print container.get_text()



from distutils.core import setup

setup(
    name='reporter',
    version='0.1.1',
    author='Jeroen Janssens',
    author_email='jeroen.janssens@visualrevenue.com',
    packages=['reporter'],
    url='http://pypi.python.org/pypi/reporter/',
    license='BSD',
    description='Flexible text extraction from HTML in Python.',
    long_description=open('README.md').read(),
    install_requires=[
        "beautifulsoup4 >= 4.1.3",
        "requests >= 0.14.2",
    ],
)



import re

from bs4 import element, Comment

from core import *
from selectors import *
from actors import *
from evaluators import *

# Triggers:
(
        HTML,
        PRE_TRAVERSAL,
        EVAL_PARAGRAPH,
        EVAL_CONTAINER,
        POST_TRAVERSAL,
        NEWS_CONTAINER,
        NEWS_TEXT,
) = tuple(2**i for i in range(0,7))


class Autocue(list):

    def execute(self, tag, trigger):

        if isinstance(tag, element.Tag):
            # apply to tag

            for (s, t) in self:
                if trigger & t != 0:
                    if isinstance(s, Selector):
                        s.select(tag)
                    elif isinstance(s, Actor):
                        SingleTagSelector(s).select(tag)
                    elif isinstance(s, Evaluator):
                        SingleTagSelector(Scorer(s)).select(tag)
                    else:
                        pass

        else:
            # apply to text

            text = tag
            for (s, t) in self:
                if trigger & t != 0:
                    text = s.act(text)
            return text




# An Autocue is a list of tuples ( [[[Selector] Actor] Evaluator] , Trigger )
default_autocue = Autocue()


# Split a paragraph with two consecutive line breaks into two paragraphs
default_autocue.append((RegExReplacer(pattern='<br */? *>[ \r\n]*<br */? *>', repl='</p><p>'), HTML))


# Get rid of scripts, styles and comments
default_autocue.append((CSSSelector("script", Pruner()), PRE_TRAVERSAL))
default_autocue.append((CSSSelector("body > style", Pruner()), PRE_TRAVERSAL))
default_autocue.append((CommentSelector(Pruner()), PRE_TRAVERSAL))

default_autocue.append((CSSSelector("footer", Pruner()), PRE_TRAVERSAL))
default_autocue.append((CSSSelector("aside", Pruner()), PRE_TRAVERSAL))

default_autocue.append((CSSSelector("div#comments", Pruner()), PRE_TRAVERSAL))
default_autocue.append((CSSSelector("div.blogComments", Pruner()), PRE_TRAVERSAL))
default_autocue.append((CSSSelector("div.reacties", Pruner()), PRE_TRAVERSAL))
default_autocue.append((CSSSelector("div#ngn-footer", Pruner()), PRE_TRAVERSAL))
default_autocue.append((CSSSelector("div#section-footer-fn", Pruner()), PRE_TRAVERSAL))
default_autocue.append((CSSSelector("div#footer2010", Pruner()), PRE_TRAVERSAL))

#default_autocue.append((CSSSelector("footer > *", Scorer(FixedValue(-100, "footer"))), PRE_TRAVERSAL))
#default_autocue.append((BeautifulSelector(id=re.compile("credit|copyright|comment|meta|footer|footnote|foot"), actor=Scorer(FixedValue(-100, "negative_id"))), PRE_TRAVERSAL))
# Assign prior scores
#default_autocue.append((BeautifulSelector(name=re.compile("^h1|h2|post|hentry|entry|content|article$"), actor=Scorer(FixedValue(100, "positive_name"))), PRE_TRAVERSAL))
#default_autocue.append((BeautifulSelector(id=re.compile("caption|post|hentry|entry|content|article"), actor=Scorer(FixedValue(100, "positive_id"))), PRE_TRAVERSAL))
#default_autocue.append((BeautifulSelector(attrs=re.compile("caption|post|hentry|entry|content|body|article"), actor=Scorer(FixedValue(100, "positive_class"))), PRE_TRAVERSAL))
#default_autocue.append((BeautifulSelector(name=re.compile("footer|aside"), actor=Scorer(FixedValue(-100, "negative_name"))), PRE_TRAVERSAL))
#default_autocue.append((BeautifulSelector(id=re.compile("credit|copyright|comment|meta|footer|footnote|foot"), actor=Scorer(FixedValue(-100, "negative_id"))), PRE_TRAVERSAL))
#default_autocue.append((BeautifulSelector(attrs=re.compile("credit|copyright|comment|meta|footer|footnote|foot"), actor=Scorer(FixedValue(-100, "negative_class"))), PRE_TRAVERSAL))

default_autocue.append((CSSSelector("p", actor=Scorer(FixedValue(50, "paragraph"))), PRE_TRAVERSAL))
default_autocue.append((CSSSelector("h1", actor=Scorer(FixedValue(40, "header"))), PRE_TRAVERSAL))
default_autocue.append((CSSSelector("h2", actor=Scorer(FixedValue(50, "header"))), PRE_TRAVERSAL))
default_autocue.append((CSSSelector("h3", actor=Scorer(FixedValue(40, "header"))), PRE_TRAVERSAL))
default_autocue.append((CSSSelector("h4", actor=Scorer(FixedValue(30, "header"))), PRE_TRAVERSAL))

default_autocue.append((CSSSelector("div", actor=Scorer(FixedValue(-100, "div"))), PRE_TRAVERSAL))

# A paragraph is scored based on the number of words and comma's
default_autocue.append((Scorer(RegExMatcher(',', factor=10, name="comma"), reset_children=True), EVAL_PARAGRAPH))
default_autocue.append((Scorer(RegExMatcher("(\w)+(['`]\w)?", factor=2, name="word"), reset_children=True), EVAL_PARAGRAPH))
# Too many CAPITALIZED WORDS is usually a bad thing
default_autocue.append((Scorer(RegExMatcher('[A-Z]{4,}', factor=-10, name="caps")), EVAL_PARAGRAPH))
# This is a rather specific example
#default_autocue.append((Scorer(RegExMatcher('Photo credit:', factor=-50, name="photo_credit"), reset_children=True), EVAL_PARAGRAPH))
# Sum up the scores of the tags that are inside the paragraph
default_autocue.append((ScoreAggregator(start_score=0), EVAL_PARAGRAPH))

# Penalize containers
default_autocue.append((ScoreAggregator(start_score=-70, vmin=0), EVAL_CONTAINER))


# We don't like divs too much :-)
default_autocue.append((CSSSelector("div", Scorer(FixedValue(-60))), NEWS_CONTAINER))
# Get rid of any tags that have a score below -50
default_autocue.append((ScoreSelector(threshold=-50, mode="upper", actor=Pruner()), NEWS_CONTAINER))


# Put all text on one line
default_autocue.append((RegExReplacer(pattern='\s+', repl=' '), NEWS_TEXT))
# Trim left and right whitespace
default_autocue.append((RegExReplacer(pattern='^\s+|\s+$', repl=''), NEWS_TEXT))



import re

import core
from core import *
from selectors import *
from actors import *

class Evaluator(object):
    """Evaluates one tag"""

    def __init__(self, name="evaluator", factor=1.0, threshold=None, vmin=None, vmax=None):
        self.name = name
        self.factor = factor
        self.threshold = threshold
        self.vmin = vmin
        self.vmax = vmax

    def evaluate(self, tag):
        return None

    def process(self, value):

        value *= self.factor

        if self.threshold is not None:
            if value > self.threshold:
                if self.vmax is not None:
                    value = self.vmax
                else:
                    value = 1
            else:
                if self.vmin is not None:
                    value = self.vmin
                else:
                    value = 0

        else:
            if self.vmin is not None:
                value = max(value, self.vmin)

            if self.vmax is not None:
                value = min(value, self.vmax)

        return value



class FixedValue(Evaluator):

    def __init__(self, value=0, name="fixed_value", **kwargs):

        super(FixedValue, self).__init__(name, **kwargs)
        self.value = value

    def evaluate(self, tag):
        return self.value


class RegExMatcher(Evaluator):

    def __init__(self, pattern="\w+", name="regex_counter", **kwargs):

        super(RegExMatcher, self).__init__(name, **kwargs)

        if isinstance(pattern, str):
            self.pattern = re.compile(pattern, flags=re.UNICODE)
        else:
            self.pattern = pattern

    def evaluate(self, tag):

        text = tag.get_text()
        num_matches = len(re.findall(self.pattern, text))
        return self.process(num_matches)


class ScoreAggregator(Evaluator):

    def __init__(self, start_score=0, name="score_aggregator", **kwargs):

        super(ScoreAggregator, self).__init__(name, **kwargs)
        self.start_score = start_score

    def evaluate(self, tag):

        score = self.start_score

        for child in tag.children:
            score += core.get_total_score(child)

        return self.process(score)


class ChildrenCounter(Evaluator):

    def __init__(self, selector=None, only_direct_children=True, name="children_counter", **kwargs):

        super(ChildrenCounter, self).__init__(name, **kwargs)
        self.selector = selector
        self.points_per_child = points_per_child
        self.only_direct_children = only_direct_children

        self.selector.actor = Yielder()

    def evaluate(self, tag):

        child_count = 0
        for child in self.selector.select(tag):
            if self.only_direct_children:
                if not child in tag.children:
                    continue

            child_count += 1

        return child_count * self.points_per_child

